// Title:	BiDiPro DLL module
// Desc:	SPI flash
// Ver:		0.1.6
// Found:	2008-09-26
// Last:	2012-04-30
// CPU:		386+
// Model:	Tiny
// Compil:	Borland C++ 3.1 (DOS)
// Author:	Oleg O. Chukaev aka 0leg


/****
25F BiDiPro module
Copyright (c) 2008-2012, 㪠   <oleg.chukaev@mail.ru>

蠥 ୮ ࠭  ᯮ짮    室
, ⠪   筮 ଥ,  ﬨ  ,  ᮡ ᫥
᫮:

1.  ୮ ࠭ 室   ⠢ 㪠
    㢥  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩.
2.  ୮ ࠭ 筮   ࠭ 㪠
    ଠ  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩  㬥樨 /  㣨 ਠ, ⠢塞
    ࠭.

      /  
"  "  -  ,   
, ,    ,  
      .    ,
    ,      ,
         ,   
/   ,    ,  
,   , ,   
,      
(,     ,  , 
,    -    , 
      ),   
         .
****/


#include <stdlib.h>
#include <limits.h>
#include "../dlldata.h"
#include "../errors.h"
#include "../bidi.h"


#ifdef LINUXXX
 extern DLLDATA *dlp;
#else
# define dlp ((DLLDATA *) NULL)		/* ⥫    */
#endif

#define bdport dlp->bdprt		/* ⥫  hardware ࠩ */

typedef struct CHDEFS {
	dword size;	// 𬪮
} CHDEFS;

CHDEFS chd[]={
//	size
	{0x00000001L},	//Status Register
	{0x00000008L},	//ID Number for Excel ES25M*A, Ramtron FM25VN* chips
	{0x00000010L},	//UID for Numonyx M25PX* chips
	{0x00000040L},	//OTP Page for AMIC A25L032 chip
	{0x00000041L},	//OTP Page for Numonyx M25PX* chips
	{0x00000080L},	//OTP SR for Atmel AT25F512B chip
	{0x00000100L},	//Parameter Page for Excel ES25P* chips
	{0x00004000L},	//FM25V01
	{0x00008000L},	//FM25V[N]02
	{0x00010000L},	//512
	{0x00020000L},	//010
	{0x00040000L},	//020
	{0x00080000L},	//040
	{0x00100000L},	//080
	{0x00200000L},	//016
	{0x00400000L},	//032
};

// ७   CFG:
static struct CFGKEY cfgkeys[] = {
	{'D', "id_alg", 0, ULONG_MAX, 20},		//Read ID alg.
	{'D', "wrsr_alg", 0, ULONG_MAX, 20},		//WRSR alg.
	{'D', "er_alg", 0, ULONG_MAX, 99},		//Erase alg.
	{'D', "rd_alg", 0, ULONG_MAX, 9},		//Read alg.
	{'D', "wr_alg", 0, ULONG_MAX, 99},		//Write alg.
	{'H', "flags", 0, ULONG_MAX, 0xffffL},		//Flags
	{'D', "wpage", 0, ULONG_MAX, 65536L},		//Page size
	{'D', "wtime", 20, ULONG_MAX, 100000L},		//Write time (us)
	{'D', "etime", 100000L, ULONG_MAX, 40000000L},	//Erase time (us)
	{'h', "offset", 0, ULONG_MAX, ULONG_MAX},	//Offset of 1st byte to read
	{'h', "bc_size", 0, ULONG_MAX, ULONG_MAX},	//Chip size for blank checking
	{0, {0}, 0, 0},					//End of table
};

#define cfg_id_alg	cfgkeys[0].data
#define cfg_wrsr_alg	cfgkeys[1].data
#define cfg_er_alg	cfgkeys[2].data
#define cfg_rd_alg	cfgkeys[3].data
#define cfg_wr_alg	cfgkeys[4].data
#define cfg_flags	cfgkeys[5].data
#define cfg_wpage	cfgkeys[6].data
#define cfg_wtime	cfgkeys[7].data
#define cfg_etime	cfgkeys[8].data
#define cfg_offset	cfgkeys[9].data
#define cfg_bc_size	cfgkeys[10].data

/* Output, 1B */
#define PIN_CE		0x01
#define PIN_WP		0x04
#define PIN_SI		0x08
#define PIN_SCK		0x10
#define PIN_HOLD	0x20

/* Output, 1C */
#define PIN_SST45_WP	0x01

/* Input, 1A */
#define PIN_SO		0x01

#define SPI_READ	0x03
#define	SPI_READ_PP	0x53
#define	SPI_READ_ID_NUM	0x4b
#define	SPI_ROTP	0x48
#define	SPI_ROTP_ATMEL	0x77
#define	SPI_SNR		0xc3
#define SPI_SERASE	0x20
#define SPI_BERASE	0x52
#define SPI_CERASE_60	0x60
#define SPI_CERASE_62	0x62
#define SPI_CERASE_C7	0xc7
#define	SPI_ERASE_PP	0xd5
#define SPI_PRG		0x02
#define	SPI_PRG_PP	0x52
#define	SPI_POTP	0x42
#define	SPI_POTP_ATMEL	0x9b
#define SPI_AAI1	0xaf
#define SPI_AAI2	0xad
#define SPI_RDSR	0x05
#define SPI_EWSR	0x50
#define SPI_WREN	0x06
#define SPI_WRSR	0x01
#define SPI_WRDI	0x04
#define SPI_RDID_15	0x15
#define SPI_RDID_90	0x90
#define SPI_RDID_JEDEC	0x9f
#define SPI_RDID_AB	0xab
#define	SPI_UNPROT	0x39
#define	SPI_PPRGAE	0x82

#define WRSR_TIME	150000L		/* Delay for WRSR command */

#define	FLAGS_AT26F004	0x0001
#define	FLAGS_AT25DF	0x0002
#define	FLAGS_ADDR_16	0x0004


static char *copyright = "25F 0.1.6  SPI flash module  "
	"(c) Oleg O. Chukaev  ["__DATE__"]";
static word chtype = 0xffff;
static CHDEFS *pchd;
static byte pin_wp = 0;


/*
 | 몫祭 ⠭
 */
static void power_off (void) {
	bdport (PRST);
}


/*
 | 祭 ⠭
 */
static void power_on (void) {

	bdport (PFORCE | PWR1 | PAIN);		/* DD1:IOO */
	bdport (PFORCE | PWR2);			/* DD4:OOO */
	dlp->voltage (dlp->Vcc, 0, 0);
	bdport (PWB2 | 0x80);			/* Vcc pin 36 */
	bdport (PWA2 | BITTER);
	bdport (PWB1 | PIN_HOLD | pin_wp | PIN_CE);
	bdport (PWC1 | PIN_SST45_WP);		/* 1  ࠧ襭  */
	dlp->mcsdelay (10000);
}


/*
 | Get value of one pin
 |
 | Ret: 0 / 1
 */
static byte get_pin (byte pin) {
	return (bdport (PRA1) & pin) ? 1 : 0;
}


/*
 | Set all pins
 */
static void set_pins (byte pins) {
	bdport (PWB1 | pins);
}


static void pin_ce_1 (void) {
  	set_pins (PIN_HOLD | pin_wp | PIN_CE);
}


static void pin_ce_0 (void) {
  	set_pins (PIN_HOLD | pin_wp);
}


/*
 |  
 */
static byte spi_read_byte (void) {

	word i;
	byte d = 0;

	for (i = 0; i < 8; i++) {
		d <<= 1;
		if (cfg_rd_alg != 1) {
			/*  設⢠  訩    SO
			 * 񫪨  ᯠ ⮣   SCK,
			 *  񫪭 訩    SI.
			 */
			d |= get_pin (PIN_SO);
		}
		set_pins (pin_wp | PIN_HOLD | PIN_SCK);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);		/* SCK=0 */
		if (cfg_rd_alg == 1) {
			/*   SST45*   ⠪,   
			 * 񫪨 _᫥騬_ ᮬ  SCK.
			 */
			d |= get_pin (PIN_SO);
		}
	}
	return d;
}


/*
 | । 
 */
static void spi_write_byte (byte d) {

	word i;
	byte z;

	for (i = 0; i < 8; i++) {
		z = (d & 0x80) ? PIN_SI : 0;
		set_pins (pin_wp | PIN_HOLD | z);		/* data */
		set_pins (pin_wp | PIN_HOLD | PIN_SCK | z);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);			/* SCK=0 */
		d <<= 1;
	}
}


/*
 | । 16-  24-⭮ 
 */
static void spi_write_addr (dword addr) {

	if ((cfg_flags & FLAGS_ADDR_16) == 0)
		spi_write_byte ((addr >> 16L) & 0xff);	/* A23..A16 */
	spi_write_byte ((addr >> 8L) & 0xff);		/* A15..A8 */
	spi_write_byte ((addr >> 0L) & 0xff);		/* A7..A0 */
}


/*
 | Send Write Enable command
 |
 | CE = 1 on exit
 */
static void send_wren (void) {

	if (cfg_wr_alg != 4) {
		pin_ce_0 ();
		spi_write_byte (SPI_WREN);
		pin_ce_1 ();
	}
}


/*
 | Read Status Register
 |
 | CE = 1 on exit
 */
static byte read_status_reg (void) {

	byte d;

	pin_ce_0 ();			//!!!! s_w_b  CE  0...
	spi_write_byte (SPI_RDSR);
	d = spi_read_byte ();
	pin_ce_1 ();
	return d;
}


/*
 | Write Status Register
 |
 | CE = 1 on exit
 */
static void write_status_reg (byte d) {

	pin_ce_0 ();
	if (cfg_wrsr_alg == 0) {
		spi_write_byte (SPI_WREN);
	} else /* cfg_wrsr_alg == 1 */ {
		spi_write_byte (SPI_EWSR);
	}
	pin_ce_1 ();
	dlp->mcsdelay (1000);		//!!!!
	pin_ce_0 ();			//!!!!
	spi_write_byte (SPI_WRSR);
	spi_write_byte (d);
	pin_ce_1 ();
	dlp->mcsdelay (WRSR_TIME);
}


/*
 | Unprotect blocks for AT26F004
 */
static void unprotect_at26f004 (void) {

	int i;
	dword sa[] = {
		0x00000L, 0x10000L, 0x20000L, 0x30000L, 0x40000L, 0x50000L,
		0x60000L, 0x70000L, 0x78000L, 0x7a000L, 0x7c000L
	};

	/* Clear SPRL bit in the Status Register */
	write_status_reg (0x00);
	/* Unprotect sectors */
	for (i = 0; i < sizeof(sa) / sizeof(sa[0]); i++) {
		send_wren ();
		pin_ce_0 ();
		spi_write_byte (SPI_UNPROT);
		spi_write_addr (sa[i]);
		pin_ce_1 ();
	}
}


/*
 | Unprotect blocks
 */
static void unprotect (void) {

	if (cfg_wr_alg == 4)
		return;

	if (cfg_flags & FLAGS_AT26F004) {
		unprotect_at26f004 ();
	} else if (cfg_flags & FLAGS_AT25DF) {
		write_status_reg (0x00);	/* Clear SPRL bit */
		write_status_reg (0x00);	/* Clear bits 5:2 */
	} else {
		write_status_reg (0x00);	/* Clear BP0, BP1, ... */
	}
}


/*
 | Poll BUSY/WIP bit
 */
static int poll_busy_bit (dword timeout) {

	dword i;

	for (i = 0; i <= timeout / 20; i++) {
		dlp->mcsdelay (20);
		if ((read_status_reg () & 0x01) == 0x00) {
			return 0;
		}
	}
	return 1;
}


/*
 | Compare page
 */
static int compare_page (dword caddr, dword baddr) {

	dword i;
	byte d;

	spi_write_byte (SPI_READ);
	spi_write_addr (caddr);
	for (i = 0; i < cfg_wpage; i++) {
		d = spi_read_byte ();
		if (dlp->getbytebuf (baddr + i) != d) {
			dlp->adr = baddr + i;
			dlp->edata = d;
			return 1;
		}
	}
	return 0;
}


/*
 | Read, compare, blank check
 |
 | 0	[JEDEC]
 | 1	SST45
 | 4	ID Number (Excel ES25M*A); OTP Page (M25PX*)
 | 5	Parameter Page (Excel ES25P*)
 | 6	OTP page (AMIC A25L032)
 | 7	UID (M25PX*)
 | 8	OTP Sec. Reg. (AT25F512B)
 | 9	Unique Serial Number (FM25VN*)
 */
static int read_all (void) {

	dword adr;
	dword start_addr;
	dword chip_size;
	dword istep;
	dword errors = 0;
	dword addr_div;
	byte dat;

	if (dlp->funcode == DFN_FBLANK) {
		start_addr = 0;
		chip_size = cfg_bc_size ? cfg_bc_size : pchd->size;
	} else {
		start_addr = cfg_offset;
		chip_size = pchd->size;
	}
	istep = chip_size / 64L;
	addr_div = max (chip_size / pchd->size, 1);

	/* Status Register */
	if (pchd->size == 1) {
		power_on ();
		dat = read_status_reg ();
		power_off ();
		dlp->setbytebuf (0, dat);
		return 0;
	}

	/* Flash */
	power_on ();
	pin_ce_0 ();
	switch (cfg_rd_alg) {
        case 0:
		/* 25F* main flash array */
		spi_write_byte (SPI_READ);
		spi_write_addr (start_addr);
                break;
	case 1:
		/* SST45* */
		spi_write_byte (0xff);
		spi_write_addr (0x000000);
		spi_write_byte (0xff);
		spi_write_byte (0xff);
		break;
        case 4:
		/* ID Number (Excel ES25M*A); OTP Page (M25PX*) */
		spi_write_byte (SPI_READ_ID_NUM);
		spi_write_addr (0x000000);
		spi_write_byte (0x00);
                break;
	case 5:
		/* Parameter Page (Excel ES25P*) */
		spi_write_byte (SPI_READ_PP);
		spi_write_addr (0x000000);
                break;
	case 6:
		/* OTP page (AMIC A25L032) */
		spi_write_byte (SPI_ROTP);
		spi_write_addr (0x000000);
		spi_write_byte (0x00);
                break;
	case 7:
		/* UID (M25PX*) */
		spi_write_byte (SPI_RDID_JEDEC);
		spi_read_byte ();		/* Manufacturer = 0x20 */
		spi_read_byte ();		/* Memory type = 0x71 */
		spi_read_byte ();		/* Memory capacity = 0x?? */
		spi_read_byte ();		/* CFD length = 0x10 */
        case 8:
        	/* OTP Sec. Reg. (AT25F512B) */
		spi_write_byte (SPI_ROTP_ATMEL);
		spi_write_addr (0x000000);
		spi_write_byte (0x00);
		spi_write_byte (0x00);
                break;
	case 9:
		/* Unique Serial Number (FM25VN*) */
		spi_write_byte (SPI_SNR);
		break;
	}
	for (adr = 0; adr < chip_size; adr++) {
		dat = spi_read_byte ();
		switch (dlp->funcode) {
		case DFN_FREAD:
			dlp->setbytebuf (adr, dat);
			break;

		case DFN_FCOMPARE:
			if (dlp->getbytebuf (adr) != dat) {
				/* ⬥ ᮢ */
				dlp->setbmark (adr, 1);
				errors++;
			}
			break;

		case DFN_FBLANK:
			if (dat != 0xff) {
				dlp->adr = adr;
				power_off ();
				return ERC_BLANK;
			}
			break;
		}
		if (istep >= 4 && adr % istep == 0
				&& dlp->iprogress (adr / addr_div, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return 0;
}


/*
 | Read ID
 |
 | 0 (JEDEC):	0x9f, [read cont.,] read mfr, read type, read capacity
 |              Eon, ...
 | 1 (Atmel 1):	0x15, read mfr, read dev
 | 2 (Atmel 2):	0xab, read mfr, read dev_h, read dev_l
 | 5 (SST,Eon):	0x90, 0x00, 0x00, 0x00 / 0x01, read mfr / read dev
 |              SST, SST45, Eon
 | 7 (Elan):    0x90, 0x00, 0x00, 0x00, rd 0x7f, rd 0x7f, read mfr, read dev
 | 9 (SST):	0xab, 0x00, 0x00, 0x00, read mfr, read dev[, read mfr2]
 |              SST, [PMC], Eon (mfr == dev == ...)
 */
static int read_id (void) {

	dword id1 = 0;
	dword id2 = 0;
        int i;

	power_on ();
	pin_ce_0 ();
	switch (cfg_id_alg) {
	case 0:
		/* JEDEC ID
		 * 0x9f, [read cont.,] read mfr, read type, read capacity
		 */
		spi_write_byte (SPI_RDID_JEDEC);
                for (i = 0; i < 16; i++) {
			id1 = spi_read_byte ();
			if (id1 != 0x7f) {
				break;
			}
                }
		id2 = (spi_read_byte () << 8);
		id2 |= spi_read_byte ();
		break;

	case 1:
		spi_write_byte (SPI_RDID_15);
		id1 = spi_read_byte ();
		id2 = spi_read_byte ();
		break;

	case 2:
		spi_write_byte (SPI_RDID_AB);
		id1 = spi_read_byte ();
		id2 = (spi_read_byte () << 8L);
		id2 |= spi_read_byte ();
		break;

	case 5:
		spi_write_byte (SPI_RDID_90);
		spi_write_addr (0x000000);
		id1 = spi_read_byte ();
		pin_ce_1 ();
		pin_ce_0 ();
		spi_write_byte (SPI_RDID_90);
		spi_write_addr (0x000001);
		id2 = spi_read_byte ();
		break;

	case 7:
		spi_write_byte (SPI_RDID_90);
		spi_write_addr (0x000000);
		id1 = (spi_read_byte () << 8L);
		id1 |= spi_read_byte ();
		id2 = (spi_read_byte () << 8L);
		id2 |= spi_read_byte ();

	case 9:
		spi_write_byte (SPI_RDID_AB);
		spi_write_addr (0x000000);	/* ID addr (0: mfr_id first) */
		id1 = spi_read_byte ();
		id2 = spi_read_byte ();
		break;
	}
	power_off ();
	dlp->edata = id1 | (id2 << 16L);
	return 0;
}


/*
 | Erase
 |
 | 0 (SST 2, JEDEC):		0xC7
 | 1 (Atmel 1):			0x62
 | 2 (SST, Atmel 2):		0x60
 | 3 (SST45*):			0x60, 0x??, 0x??, 0x?? 0xD0, 0x??
 | 5 (PP on Excel ES25P*):	0xd5
 */
static int erase (void) {

	dword istep;
	dword i;
	dword w0;
	byte saved_sr;

	if (cfg_er_alg == 3) {
		pin_wp = PIN_WP;
		power_on ();
		pin_ce_0 ();
		spi_write_byte (SPI_CERASE_60);
		spi_write_byte (0x00);
		spi_write_byte (0x00);
		spi_write_byte (0x00);
		spi_write_byte (0xd0);
		spi_write_byte (0x00);
		dlp->mcsdelay (cfg_etime);		/* delay */
		power_off ();
		pin_wp = 0;
		return 0;
	}

	pin_wp = PIN_WP;
	power_on ();
	saved_sr = read_status_reg ();
	unprotect ();
	send_wren ();
	pin_ce_0 ();
	if (cfg_er_alg == 0) {
		spi_write_byte (SPI_CERASE_C7);
	} else if (cfg_er_alg == 1) {
		spi_write_byte (SPI_CERASE_62);
	} else if (cfg_er_alg == 2) {
		spi_write_byte (SPI_CERASE_60);
	} else if (cfg_er_alg == 5) {
		spi_write_byte (SPI_ERASE_PP);
	}
	pin_ce_1 ();
	if (cfg_etime == 0L) {
		/* poll BUSY bit */
		dlp->mcsdelay (10000);
		dlp->iprogress (1600, 0);		/* init, 16 s */
		istep = 1600 / 64;
		i = 0;
		do {
			dlp->mcsdelay (10000);
			w0 = read_status_reg ();
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			i++;
		} while ((w0 & 0x01) == 0x01);
		pin_ce_1 ();
	} else if (cfg_etime <= 300000L) {
		/* <= 0.3 s */
		dlp->mcsdelay (cfg_etime);		/* delay */
	} else {
		/* etime */
		dlp->iprogress (cfg_etime / 1000, 0);	/* init */
		istep = cfg_etime / 64000L;
		for (i = 0; i < cfg_etime / 1000; i++) {
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			dlp->mcsdelay (1000);		/* delay */
		}
	}
	dlp->iprogress (0, 1);				/* close */
	write_status_reg (saved_sr);
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Write
 |
 | AAI/Page mode:
 | 0:			0x02, addr2, addr1, addr0, data0, ..., data128(256)
 | 1 (SST):		0xaf, addr2, addr1, addr0, data0, WAIT, 0xaf, data1, ...
 | 2 (SST):		0xad, addr2, addr1, addr0, data0, data1, WAIT,
 |			0xad, data2, data3, ...
 | 3 (SST45):		0x10, addr2, addr1, addr0, data0, 0x??, WAIT, ...
 | 4 (AT26DF):		0x82, addr2, addr1, addr0, data0, ..., data256
 | 5 (ES25P PP):	0x52, addr2, addr1, addr0, data0, ..., data256
 | 6 (A25L032 OTP):	0x42, addr2, addr1, addr0, data0, ..., data63
 | 7 (AT25F512 OTP):	0x9b, addr2, addr1, addr0, data0, ..., data63
 | 8 (FM25*):		0x02, addr1, addr0, data0, ...
 */
static int write_all (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size / 64L;
	byte saved_sr;

	/* Status Register */
	if (pchd->size == 1) {
		pin_wp = PIN_WP;
		power_on ();
		write_status_reg (dlp->getbytebuf (0));
		power_off ();
		pin_wp = 0;
		return 0;
	}

	if (dlp->checkwr && (cfg_wr_alg == 0 || cfg_wr_alg == 4
			|| cfg_wr_alg == 5)) {
		dlp->iprogress ((dword)"Writing, verifying chip", 3);
	} else {
		dlp->iprogress ((dword)"Writing chip", 3);
	}
	dlp->iprogress (pchd->size, 0);

	/* Flash */
	pin_wp = PIN_WP;
	power_on ();
	saved_sr = read_status_reg ();
	unprotect ();
       	send_wren ();
	pin_ce_0 ();
	switch (cfg_wr_alg) {
	case 0:
	case 4:
	case 5:
	case 6:
        case 7:
		for (adr = 0; adr < pchd->size; adr += cfg_wpage) {
			if (cfg_wr_alg == 0) {
				spi_write_byte (SPI_PRG);
			} else if (cfg_wr_alg == 4) {
				spi_write_byte (SPI_PPRGAE);
			} else if (cfg_wr_alg == 5) {
				spi_write_byte (SPI_PRG_PP);
			} else if (cfg_wr_alg == 6) {
				spi_write_byte (SPI_POTP);
			} else if (cfg_wr_alg == 7) {
				spi_write_byte (SPI_POTP_ATMEL);
			}
			spi_write_addr (adr + cfg_offset);
			for (i = 0; i < cfg_wpage; i++) {
				spi_write_byte (dlp->getbytebuf (adr + i));
			}
			pin_ce_1 ();
//			dlp->mcsdelay (cfg_wtime * cfg_wpage);
			if (poll_busy_bit (cfg_wtime * cfg_wpage)) {
				write_status_reg (saved_sr);	//!!!
				power_off ();
				pin_wp = 0;
				dlp->adr = adr;
				return ERC_WRITE;
			}
			if (dlp->checkwr) {
				if (compare_page (adr + cfg_offset, adr)) {
					write_status_reg (saved_sr);	//!!!
					power_off ();
					pin_wp = 0;
					return ERC_WRITE;
				}
				pin_ce_1 ();
			}
			send_wren ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
                       	/* Don't write bytes 64..127 of AT25F512B OTP SR. */
			if (cfg_wr_alg == 7) {
				break;
                        }
		}
		break;

	/*  wr_alg = 1 .. 3 ६   <= 20 ,  ⢥
	 * , 祬 ॡ   । 砭  
	 *  ⥭ ॣ ﭨ. ⮬ ᯮ㥬 䨪஢
	 * প.
	 */
	case 1:
		spi_write_byte (SPI_AAI1);
		spi_write_addr (cfg_offset);
		spi_write_byte (dlp->getbytebuf (0));
		pin_ce_1 ();
		dlp->mcsdelay (cfg_wtime);
		pin_ce_0 ();
		for (adr = 1; adr < pchd->size; adr++) {
			spi_write_byte (SPI_AAI1);
			spi_write_byte (dlp->getbytebuf (adr));
			pin_ce_1 ();
			dlp->mcsdelay (cfg_wtime);
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 2:
		spi_write_byte (SPI_AAI2);
		spi_write_addr (cfg_offset);
		spi_write_byte (dlp->getbytebuf (0));
		spi_write_byte (dlp->getbytebuf (1));
		pin_ce_1 ();
		dlp->mcsdelay (cfg_wtime);
		pin_ce_0 ();
		for (adr = 2; adr < pchd->size; adr += 2) {
			spi_write_byte (SPI_AAI2);
			spi_write_byte (dlp->getbytebuf (adr));
			spi_write_byte (dlp->getbytebuf (adr+1));
			pin_ce_1 ();
			dlp->mcsdelay (cfg_wtime);
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 3:
		for (adr = 0; adr < pchd->size; adr++) {
			spi_write_byte (0x10);
			spi_write_addr (adr + cfg_offset);
			spi_write_byte (dlp->getbytebuf (adr));
			spi_write_byte (0xff);
			dlp->mcsdelay (cfg_wtime);
			pin_ce_1 ();
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

        case 8:
		spi_write_byte (SPI_PRG);
		spi_write_addr (adr);
		for (adr = 0; adr < pchd->size; adr++) {
			spi_write_byte (dlp->getbytebuf (adr));
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		pin_ce_1 ();
		pin_ce_0 ();
                send_wren ();		//  ?..
		break;
 	}

	/*  SST45*, FM25*   AT26DF* WRDI  ॡ */
	if (cfg_wr_alg != 3 && cfg_wr_alg != 4 && cfg_wr_alg != 8) {
		spi_write_byte (SPI_WRDI);
	}
	write_status_reg (saved_sr);
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Test
 */
static int test (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	w1 = read_status_reg ();
	unprotect ();
	w2 = read_status_reg ();
	send_wren ();
	w3 = read_status_reg ();

	pin_ce_1 ();
	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return 0;
}


/*
 | Lock
 */
static int lock_chip (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	w1 = read_status_reg ();
	write_status_reg (0x3c);
	w2 = read_status_reg ();
	send_wren ();
	w3 = read_status_reg ();

	pin_ce_1 ();
	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return 0;
}


/*
 | ⠭ ⨯ ROM
 */
static word settype (void) {

	word i, en;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize) {
			en = chd[i].size > 1;	/* 1 -- size of status reg. */
			dlp->funmask[DFN_FCOMPARE] = en;
			dlp->funmask[DFN_READID] = en;
			dlp->funmask[DFN_ERASE] = en;
			dlp->funmask[DFN_FBLANK] = en;
			dlp->funmask[DFN_TEST] = en;
			dlp->funmask[DFN_LOCK] = en;
			if (cfg_wr_alg == 99)
				dlp->funmask[DFN_FWRITE] = 0;
			if (cfg_er_alg == 99)
				dlp->funmask[DFN_ERASE] = 0;
			return (chtype = i);
		}
	}
	return (chtype = 0xffff);
}


/*
 | Main code
 */
int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⠫ -樨 ⨢  settype(). */
		dlp->funmask[DFN_INIT] = 1;
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->defstate = 0xff;
		dlp->copyright = (byte *)copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		return 0;

	case DFN_INIT:
		if (settype () == 0xffff)
			return ERC_CHIPUNSP;
		pchd = &chd[chtype];
		dlp->dippins = 8;
		return 0;

	case DFN_FREAD:
		return read_all ();

	case DFN_FWRITE:
		return write_all ();

	case DFN_FCOMPARE:
		return read_all ();

	case DFN_FBLANK:
		return read_all ();

	case DFN_READID:
		return read_id ();

	case DFN_ERASE:
		return erase ();

	case DFN_TEST:
		return test ();

	case DFN_LOCK:
		return lock_chip ();

	default:
		return ERC_FUNUNSP;
	}
}

/* E0F */
